module(..., package.seeall)

require "defines"
require "story"

--local debug_show_path = true

local function set_pos(tab, pos, val)
    tab[pos.x] = tab[pos.x] or {}
    tab[pos.x][pos.y] = val
end

local function get_pos(tab, pos)
    return tab[pos.x][pos.y]
end

local function is_pos(tab, pos)
    return tab[pos.x] ~= nil and tab[pos.x][pos.y] ~= nil
end

local function merge_pos(dst, src)
    for x, yvals in pairs(src) do
        for y, val in pairs(yvals) do
            set_pos(dst, {x=x, y=y}, val)
        end
    end
end

local function tile_is_passable(pos)
    local tile = game.gettile(pos.x, pos.y)
    if tile.collideswith("player-layer") then return false end
    local turrets = game.findentitiesfiltered{area={{pos.x, pos.y}, {pos.x+1, pos.y+1}}, type="electric-turret"}
    if #turrets > 0 then return false end
    local walls = game.findentitiesfiltered{area={{pos.x, pos.y}, {pos.x+1, pos.y+1}}, type="wall"}
    if #walls > 0 then return false end
    return true
end

local function find_path(start, goal)
    --TODO be somehow smarter about managing the open set
    local function h(a, b)
        return math.abs(a.x-b.x) + math.abs(a.y-b.y)
    end
    local function contains(table, elem)
        for _, v in ipairs(table) do --FIXME this is slow
            if v.x == elem.x and v.y == elem.y then return true end
        end
        return false
    end

    start.x = math.floor(start.x)
    start.y = math.floor(start.y)
    goal.x = math.floor(goal.x)
    goal.y = math.floor(goal.y)

    --print("start " .. start.x .. " x " .. start.y)
    --print("goal " .. goal.x .. " x " .. goal.y)
    local closed = {}
    local open = {start}
    local came_from = {}
    local g_score = {}
    set_pos(g_score, start, 0)
    local f_score = {}
    set_pos(f_score, start, get_pos(g_score, start) + h(start, goal))

    while #open > 0 do
        local cur = table.remove(open) --this is the one with the lowest f_score
        if cur.x == goal.x and cur.y == goal.y then
            local found_path = {}
            set_pos(found_path, cur, 1)
            while is_pos(came_from, cur) do
                cur = get_pos(came_from, cur)
                set_pos(found_path, cur, 1)
            end
            return found_path
        end
        set_pos(closed, cur, 1)
        for _, dir in pairs({{0, -1}, {0, 1}, {-1, 0}, {1, 0}}) do
            local new_pos = { x = cur.x+dir[1], y = cur.y+dir[2] }
            if not is_pos(closed, new_pos) and tile_is_passable(new_pos) then --could use a continue here
                local tentative = get_pos(g_score, cur) + 1
                local is_in_open = contains(open, new_pos)
                if not is_in_open or tentative < get_pos(g_score, new_pos) then
                    set_pos(came_from, new_pos, cur)
                    set_pos(g_score, new_pos, tentative)
                    set_pos(f_score, new_pos, tentative + h(new_pos, goal))
                    if not is_in_open then
                        --keep open sorted so we can just take from the back
                        local insert_pos = #open+1
                        for k, v in ipairs(open) do --FIXME this is slow
                            if get_pos(f_score, new_pos) > get_pos(f_score, v) then
                                insert_pos = k
                                break
                            end
                        end
                        table.insert(open, insert_pos, new_pos) --FIXME this is slow
                    end
                end
            end
        end
    end
    return nil
end

local function place_copper(path)
    for x, v in pairs(path) do
        for y, _ in pairs(v) do
            game.createentity{name="copper-ore", position={x,y}}
        end
    end
end

TDStory = {}

function TDStory:init()
    game.alwaysday = true
    game.mapsettings.enemy_expansion.enabled = false
    game.evolutionfactor = 0
    game.mapsettings.enemy_evolution.enabled = false
    game.player.force.resetrecipes()
    game.player.force.resettechnologies()
    game.player.force.disableallprototypes()
    game.player.disablerecipegroups()
    game.player.minimapenabled = false

    local character = game.player.character
    game.player.character = nil
    if character then character.destroy() end

    for _, v in ipairs(self.tdcontext.config.enabled_recipes) do
        game.player.force.recipes[v].enabled = true
    end

    for _, v in ipairs(self.tdcontext.config.enabled_technologies) do
        game.player.force.technologies[v].enabled = true
    end

    for _, v in ipairs(self.tdcontext.config.starting_equipment) do
        game.player.insert(v)
    end

    for _, v in ipairs(self.tdcontext.config.spawn_positions) do
        local path = find_path(v, self.tdcontext.config.target_pos)
        assert(path)
        merge_pos(obj.tdcontext.path, path)
    end

    if debug_show_path then place_copper(obj.tdcontext.path) end

    self:update_goal()
end

function TDStory:new(tdconfig)
    table.insert(tdconfig.storytable[1], { action = function() TDStory.start_timer(obj) end })
    table.insert(tdconfig.storytable[1], { condition = function() return TDStory.waves_over(obj) end })

    --Inheritance in Lua. I have no idea what I'm doing.
    obj = story.Story:new(tdconfig.storytable)
    setmetatable(self, {__index = story.Story})
    setmetatable(obj, self)
    self.__index = self

    obj.tdcontext = {}
    obj.tdcontext.hp = tdconfig.hp
    obj.tdcontext.config = tdconfig
    obj.tdcontext.cur_wave = 0
    obj.tdcontext.timer = -1
    obj.tdcontext.path = {}
    return obj
end

function TDStory:enable_next_wave_button()
    game.player.gui.left.add{type="button", name="next-wave", caption={"next-wave"}}
end

function TDStory:start_timer()
    assert(self.tdcontext.timer == -1)
    assert(self.tdcontext.cur_wave == 0)
    self.tdcontext.timer = self.tdcontext.config.waves[1].delay * 60
    self:update_goal()
end

function TDStory:next_wave()
    self.tdcontext.cur_wave = self.tdcontext.cur_wave + 1

    if self.tdcontext.cur_wave > #self.tdcontext.config.waves then return end

    if self.tdcontext.cur_wave == #self.tdcontext.config.waves
    and game.player.gui.left["next-wave"] then
        game.player.gui.left["next-wave"].destroy()
    end

    for _, v in ipairs(self.tdcontext.config.waves[self.tdcontext.cur_wave].spawns) do
        for i = 1, v.num do
            local enemy_pos = game.findnoncollidingposition(v.type,
                                                            self.tdcontext.config.spawn_positions[v.pos],
                                                            0, 0.1)
            local enemy = game.createentity{name=v.type, position=enemy_pos}
            enemy.setcommand{type=defines.command.gotolocation,
                             destination=self.tdcontext.config.target_pos,
                             radius=0,
                             distraction=defines.distraction.none}
        end
    end

    if self.tdcontext.cur_wave < #self.tdcontext.config.waves then
        self.tdcontext.timer = self.tdcontext.config.waves[self.tdcontext.cur_wave+1].delay * 60
    else
        self.tdcontext.timer = 0
    end

    self:update_goal(true)
end

function TDStory:waves_over()
    local enemies = game.findenemyunits({x=0, y=0}, 1000) --this isn't very nice
    return self.tdcontext.cur_wave >= #self.tdcontext.config.waves and #enemies == 0
end

function TDStory:update_goal(blink)
    local time = "--"
    if self.tdcontext.timer > 0 then
        time = string.format("%d:%.2d", self.tdcontext.timer/3600, self.tdcontext.timer/60%60)
    end
    game.player.setgoaldescription({"goal-description",
                                    self.tdcontext.hp,
                                    self.tdcontext.cur_wave,
                                    #self.tdcontext.config.waves,
                                    time}, not blink)
end

function TDStory:ontick(event)
    if self.tdcontext.timer > 0 then
        self.tdcontext.timer = self.tdcontext.timer - 1
        if self.tdcontext.timer % 60 == 0 then
            self:update_goal()
        end
        if self.tdcontext.timer == 0 then
            self:next_wave()
        end
    end

    local enemies = game.findenemyunits(self.tdcontext.config.target_pos, 3)
    if #enemies > 0 then
        self.tdcontext.hp = self.tdcontext.hp - #enemies
        self:update_goal()
    end
    if self.tdcontext.hp <= 0 then
        game.setgamestate{gamefinished=true, playerwon=false}
    end
    for _, v in ipairs(enemies) do v.destroy() end
end

function TDStory:onbuiltentity(event)
    local entity = event.createdentity
    entity.destructible=false
    local pos = entity.position

    if entity.name == "stone-wall" then --turrets are already unminable
        entity.minable = nil
    end

    local rr = self.tdcontext.config.res_rect
    local in_res_rect = pos.x >= rr[1].x and pos.y >= rr[1].y
                    and pos.x < rr[2].x and pos.y < rr[2].y
    local is_res_building = not string.match(entity.name, "^turret") and entity.name ~= "stone-wall"

    local canbuild = in_res_rect == is_res_building

    pos.x = math.floor(pos.x)
    pos.y = math.floor(pos.y)
    local tile = game.gettile(pos.x, pos.y)
    if tile.name == "sand" or tile.name == "sand-dark" then canbuild = false end

    if not canbuild then
        game.player.print({"cant-build-here"})
        game.player.insert{name=entity.name, count=1}
        entity.destroy()
        return
    end

    if is_res_building then return end

    --keep previous path around, recalc only if path is affected.
    --TODO be smarter and only recalc the path that changed, not all of them
    if is_pos(self.tdcontext.path, pos) then
        new_path = {}
        for _, v in ipairs(self.tdcontext.config.spawn_positions) do
            local path = find_path(v, self.tdcontext.config.target_pos)
            if not path then
                new_path = nil
                break
            end
            merge_pos(new_path, path)
        end

        if new_path then
            self.tdcontext.path = new_path
            if debug_show_path then
                --this is ugly, but only for debugging. so, meh.
                copper = game.findentitiesfiltered{area={{-1000, -1000}, {1000, 1000}}, name="copper-ore"}
                for _, v in pairs(copper) do v.destroy() end
                place_copper(self.tdcontext.path)
            end
        else
            game.player.print({"would-block"})
            entity.die()
        end
    end
end

function TDStory:update(event, nextlevel, onwin)
    if event.name == defines.events.ontick then self:ontick()
    elseif event.name == defines.events.onbuiltentity then self:onbuiltentity(event)
    elseif event.name == defines.events.onguiclick then
        if event.element.name == "next-wave" then self:next_wave() end
    end
    story.Story.update(self, event, nextlevel, onwin)
end

function TDStory:import(context)
    self.tdcontext = context
    story.Story.import(self, context.storycontext)
end

function TDStory:export()
    self.tdcontext.storycontext = story.Story.export(self)
    return self.tdcontext
end
